/*
 * Copyright (C) 2012, H&D Wireless AB All rights reserved.
 *
 * The license to use this software in whole and in part and to
 * redistribute it in any form follows with the WiFi HW module from H&D
 * Wireless and is granted under the following restrictions:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. The name of H&D Wireless AB may not be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * 4. The software may only be used together with hardware from H&D
 *    Wireless all other use is prohibited.
 *
 * 5. The license to use and redistribute the software is granted
 *    together with the purchase of a hardware platform on a one to one
 *    basis.
 *
 * 6. The binary code may not be reversed engineered or by other means
 *    copied to circumvent this license.
 * 
 * THIS SOFTWARE IS PROVIDED BY H&D WIRELESS AB ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE EXPRESSLY AND SPECIFICALLY DISCLAIMED. IN NO EVENT
 * SHALL HD WIRELESS AB BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * For more information regarding this software license Contact H&D
 * Wireless AB (support@hd-wireless.se).
 *
 */

#ifndef OWL_PORT_H
#define OWL_PORT_H

#include <avr/io.h>
#include <ioport.h>
#include "board.h"

#include <owl/drivers/owl_board.h>
#if OWL_CONFIG_MTD
# include <owl/util/mtd/mtd.h>
#endif

void owl_putc(char c);

/**
 * Board-specific data
 *
 *
 */
struct owlboard_data {
#if OWL_CONFIG_MTD
        struct mtd_dev *mtd; /* primary nvram interface */
#endif
};


#define vsniprintf vsnprintf
#define sniprintf snprintf
#define siscanf sscanf

#define SPB800P 20 /* 1-11 used by atmel boards */

#if BOARD == XMEGA_A1_XPLAINED
# define F_CPU 32000000UL
# include <util/delay.h>

/* On the xmega_a1_xplained the USB serial-port
   is not reliable, so use another port for
   the shell */
#define OWL_BOARD_SHELL_UART                    ((void *) &USARTD0)
#define OWL_BOARD_DEBUG_UART                    OWL_BOARD_SHELL_UART

/* SPB800P interface */
#define OWL_BOARD_SPB800P_UART                  ((void *) &USARTF0)
#define OWL_BOARD_SPB800P_HOST_ATTENTION_PIN    IOPORT_CREATE_PIN(PORTF,5)

/* NOTE: The WAKEUP-PIN does not work with older SPB800E */
#if 1
# define OWL_BOARD_SPB800P_WAKEUP_PIN            IOPORT_CREATE_PIN(PORTF,6)
#endif
#define OWL_BOARD_SPB800P_SHUTDOWN_PIN          IOPORT_CREATE_PIN(PORTF,1)

/* LEDs */
#define OWL_BOARD_LED0                          LED0_GPIO
#define OWL_BOARD_LED1                          LED1_GPIO
#define OWL_BOARD_LED2                          LED2_GPIO
#define OWL_BOARD_LED3                          LED3_GPIO
#define OWL_BOARD_LED4                          LED4_GPIO
#define OWL_BOARD_LED5                          LED5_GPIO
#define OWL_BOARD_LED6                          LED6_GPIO
#define OWL_BOARD_LED7                          LED7_GPIO

#elif BOARD == XPLAIN

# define F_CPU 32000000UL
# include <util/delay.h>

#define OWL_BOARD_SHELL_UART                    ((void *) &USARTC0)
#define OWL_BOARD_DEBUG_UART                    OWL_BOARD_SHELL_UART

/* SPB800P interface */
#define OWL_BOARD_SPB800P_UART                  ((void *) &USARTD0)
#define OWL_BOARD_SPB800P_HOST_ATTENTION_PIN    IOPORT_CREATE_PIN(PORTD,5)

/* NOTE: The WAKEUP-PIN does not work with older SPB800E */
#if 1
# define OWL_BOARD_SPB800P_WAKEUP_PIN            IOPORT_CREATE_PIN(PORTD,6)
#endif
#define OWL_BOARD_SPB800P_SHUTDOWN_PIN          IOPORT_CREATE_PIN(PORTD,1)

/* LEDs */
#define OWL_BOARD_LED0                          LED0_GPIO
#define OWL_BOARD_LED1                          LED1_GPIO
#define OWL_BOARD_LED2                          LED2_GPIO
#define OWL_BOARD_LED3                          LED3_GPIO
#define OWL_BOARD_LED4                          LED4_GPIO
#define OWL_BOARD_LED5                          LED5_GPIO
#define OWL_BOARD_LED6                          LED6_GPIO
#define OWL_BOARD_LED7                          LED7_GPIO

#else
# error "board not supported"
#endif


#endif
