/*
 * Copyright (C) 2012, H&D Wireless AB All rights reserved.
 *
 * The license to use this software in whole and in part and to
 * redistribute it in any form follows with the WiFi HW module from H&D
 * Wireless and is granted under the following restrictions:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. The name of H&D Wireless AB may not be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * 4. The software may only be used together with hardware from H&D
 *    Wireless all other use is prohibited.
 *
 * 5. The license to use and redistribute the software is granted
 *    together with the purchase of a hardware platform on a one to one
 *    basis.
 *
 * 6. The binary code may not be reversed engineered or by other means
 *    copied to circumvent this license.
 * 
 * THIS SOFTWARE IS PROVIDED BY H&D WIRELESS AB ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE EXPRESSLY AND SPECIFICALLY DISCLAIMED. IN NO EVENT
 * SHALL HD WIRELESS AB BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * For more information regarding this software license Contact H&D
 * Wireless AB (support@hd-wireless.se).
 *
 */

#ifndef QUERYSTRING_H
#define QUERYSTRING_H


struct querystring {
        char *key;
        char *value;
};


/**
 * strtok()-style querystring parsing. The key-value pairs will be filled in
 * into the q parameter. querystring_lookup() can be used to find the value of a
 * given key.
 *
 * Note that any pointers in q will point back into the str buffer, so make
 * sure that str is not free'd until q is no longer used.
 *
 * @param str is a HTTP querystring of the format key1=value1&key2=value2 ...
 * @param q is an array of key-value mappings that will be filled in upon
 *        successful return.
 * @param num is the number of elements in q.
 * @return the number of key-value pair successfully parsed.
 *
 */
int querystring_parse(char *str, struct querystring *q, int num);

/**
 * Lookup a value in a querystring parsed by querystring_parse().
 *
 * @param key is the key to lookup
 * @param q is the key-value pairs obtained from querystring_parse()
 * @param num is the number of valid entries in q, the return value of
 *        querystring_parse().
 *
 * @return the value for key.
 *
 */
char *querystring_lookup(const char *key, struct querystring *q, int num);


#endif

