/*
 * Copyright (C) 2012, H&D Wireless AB All rights reserved.
 *
 * The license to use this software in whole and in part and to
 * redistribute it in any form follows with the WiFi HW module from H&D
 * Wireless and is granted under the following restrictions:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. The name of H&D Wireless AB may not be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * 4. The software may only be used together with hardware from H&D
 *    Wireless all other use is prohibited.
 *
 * 5. The license to use and redistribute the software is granted
 *    together with the purchase of a hardware platform on a one to one
 *    basis.
 *
 * 6. The binary code may not be reversed engineered or by other means
 *    copied to circumvent this license.
 *
 * THIS SOFTWARE IS PROVIDED BY H&D WIRELESS AB ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE EXPRESSLY AND SPECIFICALLY DISCLAIMED. IN NO EVENT
 * SHALL HD WIRELESS AB BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * For more information regarding this software license Contact H&D
 * Wireless AB (support@hd-wireless.se).
 *
 */

#define OWL_TRACE OWL_TRACE_DATA

#include <owl/core/owl_defs.h>
#include <owl/core/owl_shell.h>
#include <owl/core/owl_err.h>
#include <owl/core/owl_debug.h>
#include <owl/drivers/owl_board.h>
#include <owl/drivers/owl_uart.h>

#include <pico/wlp_shell.h>
#include "wlp_api.h"
#include "wlp_inet.h"

/*
 * Evaluate the owl pico API (wlp_api) using the wlp_shell.
 *
 * The functions provided by the wlp_api can be invoked from a command line,
 * using a one-to-one mapping between commands and api function wherever
 * possible.
 *
 *
 *
 *
 */

static void set_baudrate_f(struct owlsh_ctx *sh, int argc, const char *argv[]);
static void reset_f(struct owlsh_ctx *sh, int argc, const char *argv[]);
static void scan_f(struct owlsh_ctx *sh, int argc, const char *argv[]);
static void linkup_f(struct owlsh_ctx *sh, int argc, const char *argv[]);
static void linkdown_f(struct owlsh_ctx *sh, int argc, const char *argv[]);
static void set_mode_f(struct owlsh_ctx *sh, int argc, const char *argv[]);
static void get_mode_f(struct owlsh_ctx *sh, int argc, const char *argv[]);
static void set_ps_f(struct owlsh_ctx *sh, int argc, const char *argv[]);
static void set_led_f(struct owlsh_ctx *sh, int argc, const char *argv[]);
static void set_heartbeat_f(struct owlsh_ctx *sh, int argc, const char *argv[]);
static void set_poll_period_f(struct owlsh_ctx *sh, int argc, const char *argv[]);
static void conf_ps_f(struct owlsh_ctx *sh, int argc, const char *argv[]);
static void conf_ap_f(struct owlsh_ctx *sh, int argc, const char *argv[]);
static void set_channel_f(struct owlsh_ctx *sh, int argc, const char *argv[]);

static void get_hwaddr_f(struct owlsh_ctx *sh, int argc, const char *argv[]);
static void get_network_f(struct owlsh_ctx *sh, int argc, const char *argv[]);

#if WLP_STACK_OFFLOAD
static void set_ipaddr_f(struct owlsh_ctx *sh, int argc, const char *argv[]);
static void get_ipaddr_f(struct owlsh_ctx *sh, int argc, const char *argv[]);
static void set_dhcp_f(struct owlsh_ctx *sh, int argc, const char *argv[]);
static void get_dhcp_f(struct owlsh_ctx *sh, int argc, const char *argv[]);
static void set_dhcpd_f(struct owlsh_ctx *sh, int argc, const char *argv[]);
static void socket_f(struct owlsh_ctx *sh, int argc, const char *argv[]);
static void bind_f(struct owlsh_ctx *sh, int argc, const char *argv[]);
static void listen_f(struct owlsh_ctx *sh, int argc, const char *argv[]);
static void connect_f(struct owlsh_ctx *sh, int argc, const char *argv[]);
static void close_f(struct owlsh_ctx *sh, int argc, const char *argv[]);
static void send_f(struct owlsh_ctx *sh, int argc, const char *argv[]);
static void sendto_f(struct owlsh_ctx *sh, int argc, const char *argv[]);
static void get_peeraddr_f(struct owlsh_ctx *sh, int argc, const char *argv[]);
static void get_hostbyname_f(struct owlsh_ctx *sh, int argc,
                             const char *argv[]);
static void fw_upgrade_f(struct owlsh_ctx *sh, int argc, const char *argv[]);
#endif

static void get_version_f(struct owlsh_ctx *sh, int argc, const char *argv[]);
static void get_fw_version_f(struct owlsh_ctx *sh, int argc, const char *argv[]);
static void get_fw_caps_f(struct owlsh_ctx *sh, int argc, const char *argv[]);

#if defined(__linux__) && WLP_STACK_OFFLOAD
static void send_file_f(struct owlsh_ctx *sh, int argc, const char *argv[]);
static void recv_file_f(struct owlsh_ctx *sh, int argc, const char *argv[]);
static void recv_f(struct owlsh_ctx *sh, int argc, const char *argv[]);
static void write_fd(int fd, const char *buf, int len);
#endif /* __linux__ */

static int parse_ssid(const char *str, struct wl_ssid_t *ssid);
static int parse_int(const char *str, int *out);
static int32_t parse_int32(const char *str, int32_t *out);
#if WLP_STACK_OFFLOAD
static int parse_uint16(const char *str, uint16_t *out);

static void recv_cb(void *ctx, int sockid, int len);
static void lookup_cb(void *ctx, const struct ip_addr *ip);
static void listen_cb(void *ctx, int sockid);
static void conn_cb(void *ctx, int sockid, int connected);
#endif

const struct owlsh_cmd cmd_reset = {
        "reset", reset_f, "wlp_reset()"
};
const struct owlsh_cmd cmd_set_baudrate = {
        "set_baudrate", set_baudrate_f, "wlp_set_baudrate()"
};
const struct owlsh_cmd cmd_linkup = {
        "linkup", linkup_f, "wlp_linkup(ssid, key, auth)"
};
const struct owlsh_cmd cmd_linkdown = {
        "linkdown", linkdown_f, "wlp_linkdown()"
};
const struct owlsh_cmd cmd_set_mode = {
        "set_mode", set_mode_f, "wlp_set_mode()"
};
const struct owlsh_cmd cmd_get_mode = {
        "get_mode", get_mode_f, "wlp_get_mode()"
};
const struct owlsh_cmd cmd_set_ps = {
        "set_ps", set_ps_f, "wlp_set_ps()"
};
const struct owlsh_cmd cmd_conf_ps = {
        "conf_ps", conf_ps_f, "wlp_conf_ps()"
};
const struct owlsh_cmd cmd_conf_ap = {
        "conf_ap", conf_ap_f, "wlp_conf_ap()"
};
const struct owlsh_cmd cmd_set_channel = {
        "set_channel", set_channel_f, "wlp_set_channel()"
};
const struct owlsh_cmd cmd_get_hwaddr = {
        "get_hwaddr", get_hwaddr_f, "wlp_get_hwaddr()"
};
const struct owlsh_cmd cmd_get_network = {
        "get_network", get_network_f, "wlp_get_network()"
};

#if WLP_STACK_OFFLOAD
const struct owlsh_cmd cmd_set_ipaddr = {
        "set_ipaddr", set_ipaddr_f, "wlp_set_ipaddr(ip, mask, gw, dns)",
};
const struct owlsh_cmd cmd_get_ipaddr = {
        "get_ipaddr", get_ipaddr_f, "wlp_get_ipaddr()"
};
const struct owlsh_cmd cmd_set_dhcp = {
        "set_dhcp", set_dhcp_f, "wlp_set_dhcp(enable)"
};
const struct owlsh_cmd cmd_get_dhcp = {
        "get_dhcp", get_dhcp_f, "wlp_get_dhcp()"
};
const struct owlsh_cmd cmd_set_dhcpd = {
        "set_dhcpd", set_dhcpd_f, "wlp_set_dhcpd()"
};
const struct owlsh_cmd cmd_socket = {
        "socket", socket_f, "wlp_socket(type, proto)"
};
const struct owlsh_cmd cmd_bind = {
        "bind", bind_f, "wlp_bind(sock, ip, port)"
};
const struct owlsh_cmd cmd_listen = {
        "listen", listen_f, "wlp_listen(sockid, backlog)"
};
const struct owlsh_cmd cmd_connect = {
        "connect", connect_f, "wlp_connect(sockid, ip, port)"
};
const struct owlsh_cmd cmd_close = {
        "close", close_f, "wlp_close(sockid)"
};
const struct owlsh_cmd cmd_send = {
        "send", send_f, "wlp_send(sockid, data)"
};
const struct owlsh_cmd cmd_sendto = {
        "sendto", sendto_f, "wlp_sendto(sockid, data, ip, port)"
};
const struct owlsh_cmd cmd_get_peeraddr = {
        "get_peeraddr", get_peeraddr_f, "wlp_get_peeraddr(sockid)"
};
const struct owlsh_cmd cmd_get_hostbyname = {
        "get_hostbyname", get_hostbyname_f, "wlp_get_hostbyname(dnsname)"
};
const struct owlsh_cmd cmd_fw_upgrade = {
        "fw_upgrade", fw_upgrade_f, "wlp_fw_upgrade(method, addr, port, path)"
};
#endif /* WLP_STACK_OFFLOAD */

const struct owlsh_cmd cmd_get_version = {
        "get_version", get_version_f, "print client version"
};
const struct owlsh_cmd cmd_get_fw_version = {
        "get_fw_version", get_fw_version_f, "print firmware version"
};
const struct owlsh_cmd cmd_get_fw_caps = {
        "get_fw_caps", get_fw_caps_f, "print firmware capabilities"
};
const struct owlsh_cmd cmd_set_led = {
        "set_led", set_led_f, "wlp_set_led()"
};
const struct owlsh_cmd cmd_set_heartbeat = {
        "set_heartbeat", set_heartbeat_f, "wlp_set_heartbeat()"
};
const struct owlsh_cmd cmd_set_poll_period = {
        "set_poll_period", set_poll_period_f, "wlp_set_poll_period()",
};
const struct owlsh_cmd cmd_scan = {
        "scan", scan_f, "wlp_scan(ssid)"
};

#if defined(__linux__) && WLP_STACK_OFFLOAD
#include <sys/stat.h>

const struct owlsh_cmd cmd_send_file = {
        "send_file", send_file_f, "send_file <sockid> <file>"
};
const struct owlsh_cmd cmd_recv_file = {
        "recv_file", recv_file_f, "recv_file <sockid> <file>"
};
const struct owlsh_cmd cmd_recv = {
        "recv", recv_f, "recv <sockid>"
};

static int recv_fd = -1;
static int recv_sockid = 0;

static void
write_fd(int fd, const char *buf, int len)
{
        int fpos = 0;
        int err;
        while (fpos < len) {
                if ((err = write(fd, buf + fpos, len - fpos)) < 0) {
                        owl_err("write fail:%d", err);
                        return;
                }
                fpos += err;
        }
}
#endif /* __linux__ */

#if WLP_STACK_OFFLOAD
static void
recv_cb(void *ctx, int sockid, int len)
{
        int res;
        struct owlsh_ctx *sh = ctx;
        char buf[64];

        owlsh_printf(sh, "recv_cb sockid:%d len:%d %s\r\n", sockid, len,
                     len ? "" : "(connection lost)");

        while (len > 0) {
                int recvlen;
                if ((uint32_t) len > sizeof(buf))
                        recvlen = sizeof(buf);
                else
                        recvlen = len;

                res = wlp_recv(sockid, buf, recvlen);
                if (res < 0)
                        owl_err("fail to recv %d bytes err:%d\r\n",
                                recvlen, res);
                else if (res != len)
                        owl_info("got %d bytes, %d remains\r\n",
                                 res, len - res);

                if (res > 0) {
                        len -= res;
#ifdef __linux__
                        if (recv_fd > 0)
                                write_fd(recv_fd, buf, res);
                        else
#endif /* __linux__ */
                                owlsh_hexdump(sh, "RX", buf, res);
                }
        }
}

static void
lookup_cb(void *ctx, const struct ip_addr *ip)
{
        struct owlsh_ctx *sh = ctx;
        if (ip == NULL)
                owlsh_printf(sh, "lookup_cb: host not found\r\n");
        else
                owlsh_printf(sh, "lookup_cb ip:%s\r\n", inet_ntoa(ip));
}

static void
listen_cb(void *ctx, int sockid)
{
        struct owlsh_ctx *sh = ctx;
        int acceptid;
        owlsh_printf(sh, "listen_cb sockid:%d, accepting ... ", sockid);

        if ((acceptid = wlp_accept(sockid)) > 0) {
                wlp_set_conn_cb(acceptid, conn_cb, sh);
                wlp_set_recv_cb(acceptid, recv_cb, sh);
        }

        owlsh_perr(sh, acceptid);
}

static void
conn_cb(void *ctx, int sockid, int connected)
{
        struct owlsh_ctx *sh = ctx;
        owlsh_printf(sh, "conn_cb sockid:%d connected:%d\r\n",
                     sockid, connected);
}
#endif /* WLP_STACK_OFFLOAD */

static void
reset_f(struct owlsh_ctx *sh, int argc, const char *argv[])
{
        int res;

        (void)argc;
        (void)argv;
        res = wlp_reset();
        owlsh_perr(sh, res);
}

static void
set_baudrate_f(struct owlsh_ctx *sh, int argc, const char *argv[])
{
        int res;
        int err;
        int32_t baudrate = 0;
        int rtscts = 0;
        extern void *owluart; /* defined in main.c */

        if (argc >= 3) {
                /* set_baudrate <baudrate> <rtscts> */
                if (!parse_int32(argv[1], &baudrate))
                        goto err_parse;
                if (!parse_int(argv[2], &rtscts))
                        goto err_parse;

        } else if (argc >= 2) {
                /* set_baudrate <baudrate> */
                if (!parse_int32(argv[1], &baudrate))
                        goto err_parse;

        } else {
                goto err_parse;
        }

        res = wlp_set_baudrate(baudrate, rtscts);
        if (!res) {
                owlboard_mdelay(250);
                if ((err = owluart_init(owluart, baudrate,
                                        UART_DATABITS_8, UART_PARITY_NONE,
                                        UART_STOPBITS_1, rtscts)) < 0)
                        goto err_uart;
        }

        owlsh_perr(sh, res);
        return;

err_parse:
        owlsh_printf(sh, "usage: set_baudrate <baudrate> [rtscts]\r\n");
        return;

err_uart:
        owlsh_printf(sh, "err %d (failed to configure host uart)\r\n", err);
}

static void
linkup_f(struct owlsh_ctx *sh, int argc, const char *argv[])
{
        int res;
        struct wl_ssid_t ssid;
        const char *auth;
        owl_os_memset(&ssid, 0, sizeof(ssid));
        if (argc >= 4) {
                /* linkup <ssid> <key> <auth> */
                enum wlp_auth_mode auth_mode;

                if (!parse_ssid(argv[1], &ssid))
                        goto err_parse;
                auth = argv[3];
                if (owl_os_strcmp(auth, "wpa2") == 0)
                        auth_mode = WLP_AUTH_WPA2_PSK;
                else if (owl_os_strcmp(auth, "wpa") == 0)
                        auth_mode = WLP_AUTH_WPA_PSK;
                else if (owl_os_strcmp(auth, "wep") == 0)
                        auth_mode = WLP_AUTH_WEP;
                else
                        goto err_parse;
                res = wlp_linkup(&ssid, argv[2], auth_mode);
        } else if (argc >= 3) {
                /* linkup <ssid> <key> */
                if (!parse_ssid(argv[1], &ssid))
                        goto err_parse;
                res = wlp_linkup(&ssid, argv[2], WLP_AUTH_AUTO);
        } else if (argc >= 2) {
                /* linkup <ssid> */
                if (!parse_ssid(argv[1], &ssid))
                        goto err_parse;

                res = wlp_linkup(&ssid, NULL, WLP_AUTH_AUTO);
        } else if (argc >= 1) {
                /* linkup */
                res = wlp_linkup(NULL, NULL, WLP_AUTH_AUTO);
        } else {
                goto err_parse;
        }

        owlsh_perr(sh, res);
        return;

err_parse:
        owlsh_printf(sh, "usage: linkup [ssid] [key] [wep|wpa|wpa2]\r\n");
}

static void
linkdown_f(struct owlsh_ctx *sh, int argc, const char *argv[])
{
        int res;

        (void)argc;
        (void)argv;
        res = wlp_linkdown();
        owlsh_perr(sh, res);
}

static void
set_mode_f(struct owlsh_ctx *sh, int argc, const char *argv[])
{
        int res;
        uint8_t mode;

        if (argc < 2)
                goto err_parse;
        if (!owl_os_strcmp(argv[1], "sta"))
                mode = WL_MODE_STA;
        else if (!owl_os_strcmp(argv[1], "ap"))
                mode = WL_MODE_AP;
        else if (!owl_os_strcmp(argv[1], "adhoc"))
                mode = WL_MODE_ADHOC;
        else if (!owl_os_strcmp(argv[1], "wps"))
                mode = WL_MODE_WPS;
        else
                goto err_parse;

        res = wlp_set_mode(mode);
        owlsh_perr(sh, res);
        return;

err_parse:
        owlsh_printf(sh, "usage: set_mode <sta|ap|adhoc|wps>\r\n");

}


static void
get_mode_f(struct owlsh_ctx *sh, int argc, const char *argv[])
{
        int res;
        uint8_t mode;
        const char *modestr = NULL;

        (void)argc;
        (void)argv;

        res = wlp_get_mode(&mode); /* always success */

        switch (mode) {
        case WL_MODE_STA: modestr = "sta"; break;
        case WL_MODE_AP: modestr = "ap"; break;
        case WL_MODE_ADHOC: modestr = "adhoc"; break;
        case WL_MODE_WPS: modestr = "wps"; break;
        default: break;
        };

        owlsh_printf(sh, "%s %d (%s)\r\n",
                     res >= 0 ? "ok" : "err", res,
                     res >= 0 ? modestr : "N/A");
}


static void
set_channel_f(struct owlsh_ctx *sh, int argc, const char *argv[])
{
        int res;
        int channel;

        if (argc < 2)
                goto err_parse;
        if (!parse_int(argv[1], &channel))
                goto err_parse;

        res = wlp_set_channel(channel);
        owlsh_perr(sh, res);
        return;

err_parse:
        owlsh_printf(sh, "usage: set_channel <channel>\r\n");

}

static void
set_ps_f(struct owlsh_ctx *sh, int argc, const char *argv[])
{
        int res;
        int enable;

        if (argc < 2)
                goto err_parse;
        if (!parse_int(argv[1], &enable))
                goto err_parse;
        res = wlp_set_ps(enable);

        owlsh_perr(sh, res);
        return;

err_parse:
        owlsh_printf(sh, "usage: set_ps <0|1>\r\n");

}

static void
set_led_f(struct owlsh_ctx *sh, int argc, const char *argv[])
{
        int res;
        int enable;

        if (argc < 2)
                goto err_parse;
        if (!parse_int(argv[1], &enable))
                goto err_parse;
        res = wlp_set_led(enable);

        owlsh_perr(sh, res);
        return;

err_parse:
        owlsh_printf(sh, "usage: set_led <0|1>\r\n");

}

static void
set_heartbeat_f(struct owlsh_ctx *sh, int argc, const char *argv[])
{
        int res;
        int32_t period;

        if (argc < 2)
                goto err_parse;
        if (!parse_int32(argv[1], &period))
                goto err_parse;
        res = wlp_set_heartbeat((uint32_t)period);
        owlsh_printf(sh, "period %d\r\n", (int) period);
        owlsh_perr(sh, res);
        return;

err_parse:
        owlsh_printf(sh, "usage: set_heartbeat <0|1>\r\n");

}

static void
set_poll_period_f(struct owlsh_ctx *sh, int argc, const char *argv[])
{
        int res;
        int32_t period;

        if (argc < 2)
                goto err_parse;
        if (!parse_int32(argv[1], &period))
                goto err_parse;
        res = wlp_set_poll_period((uint32_t)period);

        owlsh_perr(sh, res);
        return;

err_parse:
        owlsh_printf(sh, "usage: set_poll_period <period in ms>\r\n");

}

static void
conf_ps_f(struct owlsh_ctx *sh, int argc, const char *argv[])
{
        int res;
        int32_t ps_traffic_timeout, ps_delay, listen_interval;
        int use_ps_poll, rx_all_dtim;

        if (argc != 6)
                goto err_parse;

        if (!parse_int(argv[1], &use_ps_poll))
                goto err_parse;
        if (!parse_int32(argv[2], &ps_traffic_timeout))
                goto err_parse;
        if (!parse_int32(argv[3], &ps_delay))
                goto err_parse;
        if (!parse_int(argv[4], &rx_all_dtim))
                goto err_parse;
        if (!parse_int32(argv[5], &listen_interval))
                goto err_parse;

        owlsh_printf(sh, "X %d %d %d %d %d", use_ps_poll,
                     (int)ps_traffic_timeout,
                     (int)ps_delay,
                     rx_all_dtim,
                     (int)listen_interval);
        res = wlp_conf_ps((uint8_t)use_ps_poll,
                          (uint32_t)ps_traffic_timeout,
                          (uint32_t)ps_delay,
                          (uint8_t)rx_all_dtim,
                          (uint32_t)listen_interval);

        owlsh_perr(sh, res);
        return;

err_parse:
        owlsh_printf(sh, "usage: conf_ps <use_ps_poll> <ps_traffic_timeout>\r\n");
        owlsh_printf(sh, "               <ps_delay> <rx_all_dtim> <listen_interval>\r\n");

}

static void
conf_ap_f(struct owlsh_ctx *sh, int argc, const char *argv[])
{
        int res;

        int hidden;
        int beacon_intvl;
        int inact_tmo;
        const char *country = NULL;

        if (argc < 4)
                goto err_parse;

        if (!parse_int(argv[1], &hidden))
                goto err_parse;
        if (!parse_int(argv[2], &beacon_intvl))
                goto err_parse;
        if (!parse_int(argv[3], &inact_tmo))
                goto err_parse;
        if (argc > 4)
                country = argv[4];

        res = wlp_conf_ap(hidden, beacon_intvl, inact_tmo, country);

        owlsh_perr(sh, res);
        return;

err_parse:
        owlsh_printf(sh, "usage: conf_ap <hidden> <beacon_intvl>\r\n");
        owlsh_printf(sh, "               <inact_tmo> [country]>\r\n");

}

static void
get_hwaddr_f(struct owlsh_ctx *sh, int argc, const char *argv[])
{
        int res;
        struct wl_mac_addr_t hwaddr;

        (void)argc;
        (void)argv;

        res = wlp_get_hwaddr(&hwaddr);

        /* out vars always valid */
        owlsh_printf(sh, "%s %d (%02x:%02x:%02x:%02x:%02x:%02x)\r\n",
                     res >= 0 ? "ok" : "err", res,
                     hwaddr.octet[0], hwaddr.octet[1], hwaddr.octet[2],
                     hwaddr.octet[3], hwaddr.octet[4], hwaddr.octet[5]);
}

static void
get_network_f(struct owlsh_ctx *sh, int argc, const char *argv[])
{
        int res;
        struct wlp_network_t net;
        char ssid_str[WL_SSID_MAX_LENGTH + 1];

        (void)argc;
        (void)argv;

        res = wlp_get_network(&net);

        owlsh_printf(sh, "%s %d", res >= 0 ? "ok" : "err", res);

        /* out vars always invalid if return < 0 */
        if (res) {
                owlsh_printf(sh, "\r\n");
                return;
        }
        
        owl_os_memcpy(ssid_str, net.ssid.ssid, net.ssid.len);
        ssid_str[net.ssid.len] = 0;

        owlsh_printf(sh,
                     " (%s %02x:%02x:%02x:%02x:%02x:%02x ",
                     net.ssid.len ? ssid_str : "N/A",
                     net.bssid.octet[0], net.bssid.octet[1],
                     net.bssid.octet[2], net.bssid.octet[3],
                     net.bssid.octet[4], net.bssid.octet[5]);

        if (net.rssi != WL_RSSI_UNKNOWN)
                owlsh_printf(sh, "rssi:%d ", net.rssi);
        else
                owlsh_printf(sh, "rssi:N/A ");

        if (net.snr != WL_SNR_UNKNOWN)
                owlsh_printf(sh, "snr:%d", net.snr);
        else
                owlsh_printf(sh, "snr:N/A");

        owlsh_printf(sh, ")\r\n");
}

#if WLP_STACK_OFFLOAD
static void
set_ipaddr_f(struct owlsh_ctx *sh, int argc, const char *argv[])
{
        int res;
        struct ip_addr ip, mask, gw, dns;

        mask = ip_addr_any;
        gw = ip_addr_any;
        dns = ip_addr_any;

        if (argc < 2)
                goto err_parse;
        if (!inet_aton(argv[1], &ip))
                goto err_parse;
        if (argc >= 3 && !inet_aton(argv[2], &mask))
                goto err_parse;
        if (argc >= 4 && !inet_aton(argv[3], &gw))
                goto err_parse;
        if (argc >= 5 && !inet_aton(argv[4], &dns))
                goto err_parse;

        res = wlp_set_ipaddr(&ip, &mask, &gw, &dns);
        owlsh_perr(sh, res);
        return;

err_parse:
        owlsh_printf(sh, "usage: set_ipaddr <ip> [mask] [gw] [dns]\r\n");
}

static void
get_ipaddr_f(struct owlsh_ctx *sh, int argc, const char *argv[])
{
        int res;
        struct ip_addr ip, mask, gw, dns;

        res = wlp_get_ipaddr(&ip, &mask, &gw, &dns);
        owlsh_printf(sh, "%s %d (", res >= 0 ? "ok" : "err", res);
        owlsh_printf(sh, "ip:%s ", inet_ntoa(&ip));
        owlsh_printf(sh, "mask:%s ", inet_ntoa(&mask));
        owlsh_printf(sh, "gw:%s ", inet_ntoa(&gw));
        owlsh_printf(sh, "dns:%s)\r\n", inet_ntoa(&dns));
}

static void
set_dhcp_f(struct owlsh_ctx *sh, int argc, const char *argv[])
{
        int res;
        int dhcp;

        if (argc < 2)
                goto err_parse;
        if (!parse_int(argv[1], &dhcp))
                goto err_parse;

        res = wlp_set_dhcp(dhcp);
        owlsh_perr(sh, res);
        return;

err_parse:
        owlsh_printf(sh, "usage: set_dhcp <0|1>\r\n");
}

static void
get_dhcp_f(struct owlsh_ctx *sh, int argc, const char *argv[])
{
        int res;
        int enabled;

        res = wlp_get_dhcp(&enabled);
        owlsh_printf(sh, "%s %d (enabled:%d)\r\n",
                     res >= 0 ? "ok" : "err", res, enabled);
}

static void
set_dhcpd_f(struct owlsh_ctx *sh, int argc, const char *argv[])
{
        int res;
        int enable;

        if (argc < 2)
                goto err_parse;
        if (!parse_int(argv[1], &enable))
                goto err_parse;

        res = wlp_set_dhcpd(enable);
        owlsh_perr(sh, res);
        return;

err_parse:
        owlsh_printf(sh, "usage: set_dhcpd <0|1>\r\n");
}

static void
socket_f(struct owlsh_ctx *sh, int argc, const char *argv[])
{
        int sockid;
        int type;
        int protocol = 0;

        if (argc < 2)
                goto err_parse;

        if (!owl_os_strcmp(argv[1], "stream"))
                type = WLP_SOCK_STREAM;
        else if (!owl_os_strcmp(argv[1], "dgram"))
                type = WLP_SOCK_DGRAM;
        else if (!owl_os_strcmp(argv[1], "raw"))
                type = WLP_SOCK_RAW;
        else
                goto err_parse;

        if (argc > 2 && !parse_int(argv[2], &protocol))
                goto err_parse;

        sockid = wlp_socket(type, protocol);
        wlp_set_conn_cb(sockid, conn_cb, sh);
        wlp_set_recv_cb(sockid, recv_cb, sh);
        owlsh_perr(sh, sockid);
        return;

err_parse:
        owlsh_printf(sh, "usage: socket <stream|dgram|raw> [protocol]\r\n");

}
static void
bind_f(struct owlsh_ctx *sh, int argc, const char *argv[])
{
        int res;
        int sockid;
        struct ip_addr ip;
        uint16_t port;

        if (argc < 4)
                goto err_parse;
        if (!parse_int(argv[1], &sockid))
                goto err_parse;
        if (!inet_aton(argv[2], &ip))
                goto err_parse;
        if (!parse_uint16(argv[3], &port))
                goto err_parse;

        res = wlp_bind(sockid, &ip, port);
        owlsh_perr(sh, res);
        return;

err_parse:
        owlsh_printf(sh, "usage: bind <sockid> <ip> <port>\r\n");

}

static void
listen_f(struct owlsh_ctx *sh, int argc, const char *argv[])
{
        int res;
        int sockid;
        int backlog = 1;

        if (argc < 2)
                goto err_parse;
        if (!parse_int(argv[1], &sockid))
                goto err_parse;
        if (argc > 2 && !parse_int(argv[2], &backlog))
                goto err_parse;


        res = wlp_listen(sockid, backlog, listen_cb, sh);

        owlsh_perr(sh, res);
        return;

err_parse:
        owlsh_printf(sh, "usage: listen <sockid> [backlog]\r\n");

}

static void
connect_f(struct owlsh_ctx *sh, int argc, const char *argv[])
{
        int res;
        int sockid;
        struct ip_addr ip;
        uint16_t port;

        if (argc < 4)
                goto err_parse;
        if (!parse_int(argv[1], &sockid))
                goto err_parse;
        if (!inet_aton(argv[2], &ip))
                goto err_parse;
        if (!parse_uint16(argv[3], &port))
                goto err_parse;

        res = wlp_connect(sockid, &ip, port);

        owlsh_perr(sh, res);
        return;

err_parse:
        owlsh_printf(sh, "usage: connect <sockid> <ip> <port>\r\n");
}

static void
close_f(struct owlsh_ctx *sh, int argc, const char *argv[])
{
        int res;
        int sockid;

        if (argc < 2)
                goto err_parse;
        if (!parse_int(argv[1], &sockid))
                goto err_parse;

        res = wlp_close(sockid);

        owlsh_perr(sh, res);
        return;

err_parse:
        owlsh_printf(sh, "usage: close <sockid>\r\n");
}

static void
send_f(struct owlsh_ctx *sh, int argc, const char *argv[])
{
        int res;
        int sockid;
        char *data;
        int len;
        int pos = 0;

        if (argc < 3)
                goto err_parse;

        if (!parse_int(argv[1], &sockid))
                goto err_parse;

        if ((data = owlsh_joinargs(argc - 2, argv + 2)) == NULL) {
                res = OWL_ERR_MEM;
                goto out;
        }

        len = owl_os_strlen(data);
        owlsh_printf(sh, "sending %d bytes\r\n", len);
        while (pos < len) {

                if ((res = wlp_send(sockid, data + pos, len - pos)) < 0) {
                        free(data);
                        goto out;

                } else {
                        owlsh_printf(sh, "wrote:%d remain:%d\r\n",
                                     res, len - pos - res);
                }
                pos += res;
        }
        free(data);
        res = 0;

out:
        owlsh_perr(sh, res);
        return;

err_parse:
        owlsh_printf(sh, "usage: send <sockid> <data ...>\r\n");
}

static void
sendto_f(struct owlsh_ctx *sh, int argc, const char *argv[])
{
        int res;
        int sockid;
        struct ip_addr ip;
        uint16_t port;
        char *data;
        int len;
        int pos = 0;

        if (argc < 4)
                goto err_parse;

        if (!parse_int(argv[1], &sockid))
                goto err_parse;

        if (!inet_aton(argv[2], &ip))
                goto err_parse;

        if (!parse_uint16(argv[3], &port))
                goto err_parse;

        if ((data = owlsh_joinargs(argc - 4, argv + 4)) == NULL) {
                res = OWL_ERR_MEM;
                goto out;
        }

        len = owl_os_strlen(data);
        owlsh_printf(sh, "sending %d bytes to %s:%d\r\n", len,
                     inet_ntoa(&ip), port);

        while (pos < len) {

                if ((res = wlp_sendto(sockid, data + pos, len - pos,
                                      &ip, port)) < 0) {
                        free(data);
                        goto out;

                } else {
                        owlsh_printf(sh, "wrote:%d remain:%d\r\n",
                                     res, len - pos - res);
                }
                pos += res;
        }
        free(data);
        res = 0;

out:
        owlsh_perr(sh, res);
        return;

err_parse:
        owlsh_printf(sh, "usage: sendto <sockid> <ip> <port> <data ...>\r\n");
}

static void
get_peeraddr_f(struct owlsh_ctx *sh, int argc, const char *argv[])
{
        int res;
        int sockid;
        struct ip_addr peer;

        if (argc < 2)
                goto err_parse;
        if (!parse_int(argv[1], &sockid))
                goto err_parse;

        res = wlp_get_peeraddr(sockid, &peer);
        owlsh_perr(sh, res);
        owlsh_printf(sh, "peer:%s\r\n", inet_ntoa(&peer));
        return;

err_parse:
        owlsh_printf(sh, "usage: get_peeraddr <sockid>\r\n");
}

static void
get_hostbyname_f(struct owlsh_ctx *sh, int argc, const char *argv[])
{
        int res;

        if (argc < 2)
                goto err_parse;

        res = wlp_get_hostbyname(argv[1], lookup_cb, sh);
        owlsh_perr(sh, res);
        return;

err_parse:
        owlsh_printf(sh, "usage: get_hostbyname <dnsname>\r\n");
}

static void
fw_upgrade_f(struct owlsh_ctx *sh, int argc, const char *argv[])
{
        struct ip_addr addr;
        int res;
        uint16_t port = 8080;
        uint8_t method = WLP_UPGRADE_METHOD_HTTP_POST;
        char *path = NULL;

        wlp_get_ipaddr(&addr, NULL, NULL, NULL);

        if (argc > 5 || argc < 1)
                goto err_parse;

        if (argc > 1) {
                if (!owl_os_strcmp(argv[1], "http_post"))
                        method = WLP_UPGRADE_METHOD_HTTP_POST;
                else
                        goto err_parse;
        }

        if (argc > 2) {
                if (!inet_aton(argv[2], &addr))
                        goto err_parse;
        }

        if (argc >= 4)
                port = owl_os_atoi(argv[3]);

        if (argc >= 5)
                path = (char *)argv[4];

        res = wlp_fw_upgrade(method, &addr, port, path);
        owlsh_perr(sh, res);
        if (WLP_UPGRADE_METHOD_HTTP_POST == method)
                owlsh_printf(sh, "Listening on %s:%d/%s\r\n", inet_ntoa(&addr),
                             port, path ? path : "");
        return;

err_parse:
        owlsh_printf(sh, "usage: fw_upgrade <method, e.g. \"http_post\"> [<ip> <port> <path>]\r\n");

}
#endif /* WLP_STACK_OFFLOAD */

static void
get_version_f(struct owlsh_ctx *sh, int argc, const char *argv[])
{
        (void)argc;
        (void)argv;

        owlsh_printf(sh, "sw version:%d.%d.%d\r\n",
                     WLP_MAJOR_VERSION(WLP_VERSION_CODE),
                     WLP_MINOR_VERSION(WLP_VERSION_CODE),
                     WLP_PATCH_VERSION(WLP_VERSION_CODE));
}

static void
get_fw_version_f(struct owlsh_ctx *sh, int argc, const char *argv[])
{
        int res;
        uint32_t code;
        char build[16];
        char hw_rev[16];

        (void)argc;
        (void)argv;

        res = wlp_get_fw_version(&code, build, sizeof build,
                                 hw_rev, sizeof hw_rev);
        owlsh_perr(sh, res);
        owlsh_printf(sh, "fw version:%d.%d.%d build:%s\r\n",
                     WLP_MAJOR_VERSION(code), WLP_MINOR_VERSION(code),
                     WLP_PATCH_VERSION(code), build);
        owlsh_printf(sh, "hw revision:%s\r\n",
                     owl_os_strlen(hw_rev) ? hw_rev : "unknown");
}

static void
get_fw_caps_f(struct owlsh_ctx *sh, int argc, const char *argv[])
{
        int res;
        uint8_t caps;

        (void)argc;
        (void)argv;

        res = wlp_get_fw_caps(&caps);
        owlsh_perr(sh, res);
        owlsh_printf(sh, "caps:0x%02x\r\n", caps);
}

static void
scan_complete_cb(void *ctx, int cnt)
{
	struct owlsh_ctx *sh = ctx;
	struct wlp_network_list *netlist;
	int i;

        if (owlsh_compact(sh)) {

#if OWL_CONFIG_SHELL_OUTPUT_COMPACT
                int err;
                owl_unused(cnt);
                if ((err = wlp_get_network_list(&netlist)) < 0) {
                        owlsh_printf(sh, "\r\n");
                        return;
                }

                /* <ssid len>;<ssid>;<hwaddr>;
                   <none|wep|wpa|wpa-psk|wpa2|wpa2-psk>;<rssi> */

                for (i = 0; i < netlist->cnt; i++) {
                        struct wlp_network_t *net = &netlist->net[i];
                        char ssid_buf[WL_SSID_MAX_LENGTH + 1];
                        const char *auth = "";
                        
                        owl_os_memset(ssid_buf, 0, sizeof(ssid_buf));
                        owl_os_memcpy(ssid_buf, net->ssid.ssid, net->ssid.len);

                        switch (net->auth) {
                        case WLP_AUTH_AUTO:
                                auth = "none";
                                break;
                        case WLP_AUTH_WEP:
                                auth = "wep";
                                break;
                        case WLP_AUTH_WPA_PSK:
                                auth = "wpa-psk";
                                break;
                        case WLP_AUTH_WPA2_PSK:
                                auth = "wpa2-psk";
                                break;
                        case WLP_AUTH_WPA:
                                auth = "wpa";
                                break;
                        case WLP_AUTH_WPA2:
                                auth = "wpa2";
                                break;
                        default:
                                owl_err("invalid auth:%d", auth);
                                owl_assert(0);
                        }

                        owlsh_printf(
                                sh,
                                "%d;%s;%02x:%02x:%02x:%02x:%02x:%02x;%s;%d\r\n",
                                net->ssid.len, ssid_buf,
                                net->bssid.octet[0], net->bssid.octet[1],
                                net->bssid.octet[2], net->bssid.octet[3],
                                net->bssid.octet[4], net->bssid.octet[5],
                                auth, net->rssi);
                        
                }
                owlsh_printf(sh, "\r\n");

#endif /* OWL_CONFIG_SHELL_OUTPUT_COMPACT */
        } else  {
                
#if OWL_CONFIG_SHELL_OUTPUT_FULL
                if (cnt == 0) {
                        owlsh_printf(sh, "no nets found\r\n");
                        return;
                }

                if (wlp_get_network_list(&netlist)) {
                        owlsh_printf(sh, "failed to get network list\r\n");
                        return;
                }

                owlsh_printf(sh, "\r\n");
                owlsh_printf(sh,
                             "BSSID              TYPE   CH RSSI SNR  "
                             "AUTH      SSID\r\n");
                owlsh_printf(sh, "-----------------------------------"
                             "----------------------------------\r\n");
                for (i = 0; i < netlist->cnt; i++) {
                        struct wlp_network_t *net = &netlist->net[i];
                        char buf[WL_SSID_MAX_LENGTH + 1];

                        /* bssid */
                        owlsh_printf(sh, "%02x:%02x:%02x:%02x:%02x:%02x  ",
                                     net->bssid.octet[0], net->bssid.octet[1],
                                     net->bssid.octet[2], net->bssid.octet[3],
                                     net->bssid.octet[4], net->bssid.octet[5]);

                        /* type, channel, rssi, snr */
                        owlsh_printf(sh, "%s  %02d %02d  ",
                                     net->type == WL_CONN_TYPE_INFRA ?
                                     "INFRA" : "ADHOC",
                                     net->channel, (int) net->rssi);

                        /* snr */
                        if (net->snr != WL_SNR_UNKNOWN)
                                owlsh_printf(sh, "%02d   ", net->snr);
                        else
                                owlsh_printf(sh, "N/A  ");

                        /* auth */
                        switch (net->auth) {
                        case WLP_AUTH_AUTO:
                                owlsh_printf(sh, "NONE      ");
                                break;
                        case WLP_AUTH_WEP:
                                owlsh_printf(sh, "WEP       ");
                                break;
                        case WLP_AUTH_WPA_PSK:
                                owlsh_printf(sh, "WPA-PSK   ");
                                break;
                        case WLP_AUTH_WPA2_PSK:
                                owlsh_printf(sh, "WPA2-PSK  ");
                                break;
                        case WLP_AUTH_WPA:
                                owlsh_printf(sh, "WPA       ");
                                break;
                        case WLP_AUTH_WPA2:
                                owlsh_printf(sh, "WPA2      ");
                                break;
                        }

                        /* ssid */
                        owl_os_memset(buf, 0, sizeof buf);
                        owl_os_memcpy(buf, net->ssid.ssid, net->ssid.len);
                        owlsh_printf(sh, "%s", buf);

                        owlsh_printf(sh, "\r\n");
                }
#endif /* OWL_CONFIG_SHELL_OUTPUT_FULL */
        }
}

static void
scan_f(struct owlsh_ctx *sh, int argc,  const char *argv[])
{
        int res;
        struct wl_ssid_t ssid;

        if (argc > 1) {
		if (!parse_ssid(argv[1], &ssid))
			goto err_parse;
		res = wlp_scan(&ssid, scan_complete_cb, sh);
        } else {
		res = wlp_scan(NULL, scan_complete_cb, sh);
        }

        /* Print <ok|err>:<res>: in compact mode to tell machines that this
           is a multi-line reply */
        if (owlsh_compact(sh))
                owlsh_printf(
                        sh, "%s;%d%s\r\n",
                        res >= 0 ? "ok" : "err",
                        res,
                        res >= 0 ? ";" : "");
        else
                owlsh_perr(sh, res);
        
	return;

err_parse:
        owlsh_printf(sh, "usage: scan [ssid]\r");
}

#if defined(__linux__) && WLP_STACK_OFFLOAD
static void
send_file_f(struct owlsh_ctx *sh, int argc, const char *argv[])
{
        int sockid;
        const char *name;
        char buf[256];
        int read_err = 0;
        int send_err = 0;
        int len = 0;
        int fd;
        struct stat stat_buf;
        int file_len;
        int read_len = 0;

        if (argc < 3)
                goto err_parse;
        if (!parse_int(argv[1], &sockid))
                goto err_parse;
        name = argv[2];

        if ((fd = open(name, O_RDONLY)) < 0) {
                owlsh_printf(sh, "could not open '%s' fd:%d\r\n", name, fd);
                return;
        }

        if (stat(name, &stat_buf) < 0) {
                close(fd);
                owlsh_printf(sh, "could not stat %s\r\n", name);
                return;
        }

        file_len = stat_buf.st_size;
        owlsh_printf(sh, "sending %d bytes\r\n", file_len);

        while (read_len < file_len) {
                if ((read_err = read(fd, buf, sizeof(buf))) < 0)
                        goto out;

                read_len += read_err;
                send_err = 0;
                int buflen = read_err;
                int bufpos = 0;
                while (bufpos < buflen) {
                        if ((send_err = wlp_send(sockid, buf + bufpos,
                                                 buflen - bufpos)) < 0)
                                goto out;
                        len += send_err;
                        bufpos += send_err;
                        owlsh_printf(sh, "wlp_send: %d\r\n", send_err);
                }
        }

out:
        owlsh_printf(sh, "done -- read_err:%d send_err:%d len:%d\r\n",
                     read_err, send_err, len);
        close(fd);
        return;
err_parse:
        owlsh_printf(sh, "usage: send_file <sockid> <file>\r\n");
}

static void
recv_file_f(struct owlsh_ctx *sh, int argc, const char *argv[])
{
        const char *name;
        int sockid;

        if (argc < 3)
                goto err_parse;
        if (!parse_int(argv[1], &sockid))
                goto err_parse;
        name = argv[2];

        if (recv_fd >= 0) {
                owlsh_printf(sh, "closing %d\r\n", recv_fd);
                close(recv_fd);
        }

        if ((recv_fd = open(name, O_CREAT | O_RDWR | O_TRUNC, 0660)) < 0) {
                owlsh_printf(sh, "failed to open %s\r\n", name);
                return;
        }

        recv_sockid = sockid;
        owlsh_printf(sh, "rx data for sockid %d will be written to %s\r\n",
                     sockid, name);
        return;


err_parse:
        owlsh_printf(sh, "usage: recv_file <sockid> <file>\r\n");
}


static void
recv_f(struct owlsh_ctx *sh, int argc, const char *argv[])
{
        if (recv_fd < 0)
                return;

        owlsh_printf(sh, "closing %d\r\n", recv_fd);
        recv_sockid = 0;
        close(recv_fd);
        recv_fd = -1;
}
#endif

#if WLP_STACK_OFFLOAD
static int
parse_uint16(const char *str, uint16_t *out)
{
        char *end;
        int val;
        val = owl_os_strtol(str, &end, 0);
        if (end == str)
                return 0;

        if (val > 65535)
                return 0;

        if (val < 0)
                return 0;

        *out = val;
        return 1;
}
#endif /* WLP_STACK_OFFLOAD */

static int
parse_int(const char *str, int *out)
{
        char *end;
        *out = owl_os_strtol(str, &end, 0);
        if (end == str)
                return 0;

        return 1;
}

static int32_t
parse_int32(const char *str, int32_t *out)
{
        char *end;
        *out = owl_os_strtol(str, &end, 0);
        if (end == str)
                return 0;

        return 1;
}

static int
parse_ssid(const char *str, struct wl_ssid_t *ssid)
{
        ssid->len = owl_os_strlen(str);
        owl_os_memcpy(ssid->ssid, str, ssid->len);
        return 1;
}
